"""Wrapper functions for making lots of IFF files."""
import aggregator
from glob import glob
import re

avhrr_base_dir = '/fjord/jgs/personal/mhiley/Satellite_Output/avhrr/global/' # e.g. 2012/level2
hirs_base_dir = '/fjord/awg/HIRS/hdf/' # e.g. hirs_base_dir+'noaa-19/2012'

satid_map = dict()
satid_map[1] = ('metop-01', 'M1')
satid_map[2] = ('metop-02', 'M2')
satid_map[5] = ('noaa-05', 'TN')
satid_map[6] = ('noaa-06', 'NA')
satid_map[7] = ('noaa-07', 'NC')
satid_map[8] = ('noaa-08', 'NE')
satid_map[9] = ('noaa-09', 'NF')
satid_map[10] = ('noaa-10', 'NG')
satid_map[11] = ('noaa-11', 'NH')
satid_map[12] = ('noaa-12', 'ND')
satid_map[13] = ('noaa-13', 'NI')
satid_map[14] = ('noaa-14', 'NJ')
satid_map[15] = ('noaa-15', 'NK')
satid_map[16] = ('noaa-16', 'NL')
satid_map[17] = ('noaa-17', 'NM')
satid_map[18] = ('noaa-18', 'NN')
satid_map[19] = ('noaa-19', 'NP')


def make_date_set(file_list):
    """Make a set of HIRS or AVHRR time strings from a list of filenames."""
    from os.path import basename
    # Careful to include those e.g. GC or WI digits... can have filenames that are otherwise identitical but for this.
    DATE_REGEX = ".*(D[0-9]{2}[0-9]{3}\.S[0-9]{4}\.E[0-9]{4}\.B[0-9]{7}\.[A-Z]{2}).*.hdf"
    ret = set()
    for file in file_list:
        m = re.match(DATE_REGEX, basename(file))
        if m:
            date_str = m.group(1)
            if date_str in ret:
                print file
                raise ValueError("date already exists in set...is this a problem??")
            ret.add(date_str)
        else:
            raise ValueError("file doesn't match DATE_REGEX; this shouldn't happen")
    return ret


def make_one_iff_zara(platform, start_date_str, start_time_str, end_time_str, hirs_filepath, avhrr_filepath, out_dir):
    """Create a shell script suitable for running a single iff job on zara.

    The args to this function basically correspond with the command line arguments iff expects.
    """
    from subprocess import check_call, CalledProcessError
    from os.path import basename
    iff_installation_path = '/home/mhiley/ShellB3/'
    iff_bin = iff_installation_path+'bin'
    iff_lib = iff_installation_path+'lib'
    log_dir = '/home/mhiley/iff_logs/'
    qsub_script_dir = '/home/mhiley/iff_qsub_scripts/'

    with open(qsub_script_dir+'IFF-'+basename(avhrr_filepath)+'.sh', mode='w') as script:
        script.write("#!/bin/bash\n")
        script.write("cd $TMPDIR\n")
        script.write("export PATH=%s:$PATH\n" % iff_bin)
        script.write("export LD_LIBRARY_PATH=%s:$LD_LIBRARY_PATH\n" % iff_lib)
        script.write("cp %s $TMPDIR\n" % hirs_filepath)
        script.write("cp %s $TMPDIR\n" % avhrr_filepath)
        cmd = 'iff %s avhrr nss %s %s %s %s %s' % (platform, start_date_str, start_time_str, 
                end_time_str, "$TMPDIR/"+basename(hirs_filepath), "$TMPDIR/"+basename(avhrr_filepath))
        script.write("%s\n" % cmd)
        script.write("cp $TMPDIR/IFF*.hdf %s\n" % out_dir)
        qsub = 'qsub -sync y -S /bin/bash -q r720.q -l h_rt=00:20:00 -l vf=4G -l matlab=0 -l friendly=1 -p -200 -o %s -e %s %s' % (log_dir, log_dir, script.name)
        script.write("#%s\n" % qsub)
    try:
        print qsub
        check_call(qsub, shell=True)
    except CalledProcessError:
        print "Error calling qsub on %s" % basename(hirs_filepath)


def make_one_iff(timestr, satid, hirs_dir, avhrr_dir, zara=False):
    """Make single IFF corresponding to a given time string."""
    from datetime import datetime, timedelta
    from avhrr import AVHRR
    from subprocess import check_call, CalledProcessError
    import os
    hirs_files = glob(hirs_dir+'NSS*'+timestr+'*.hdf')
    # fix for noaa-18 2007:  there are a handful of *.gz.hdf files screwing things up
    for thing in hirs_files:
        if thing.endswith('gz.hdf'):
            hirs_files.remove(thing)
    avhrr_files = glob(avhrr_dir+'NSS*'+timestr+'*level2.hdf')
    if not (len(hirs_files) == 1 and len(avhrr_files) == 1):
        raise ValueError("Got more than one hirs or avhrr file")
    hirs = hirs_files[0]
    avhrr = avhrr_files[0]

    platform = satid_map[satid][0]
    # the dash was needed for hirs_dir but we don't want it for the output name or clavrx will break.
    platform = platform.replace("-", "")
    tag = 'nss'
    sensor = 'avhrr'

    DATE_REGEX = ".*NSS.*D([0-9]{2})([0-9]{3})\.S([0-9]{4})\.E([0-9]{4})\..*.hdf"
    m = re.match(DATE_REGEX, avhrr)
    if not m:
        raise ValueError("regex not matched for avhrr file")
    m = re.match(DATE_REGEX, hirs)
    if not m:
        raise ValueError("regex not matched for hirs file")
    yy, ddd, ssss, eeee = m.group(1), m.group(2), m.group(3), m.group(4)
    frmt = '%y%j%H%M'
    start = datetime.strptime(yy+ddd+ssss, frmt)
    end = datetime.strptime(yy+ddd+eeee, frmt)
    if start > end:
        end = end + timedelta(days=1)
    bowtie = False
    filelist = [avhrr, hirs]

    out_name = "IFF_%(platform)s_%(sensor)s_%(tag)s_d%(start_day)s_t%(start_time)s_e%(end_time)s_c%(creation_time)s.hdf"
    out_name = out_name % {"platform"      : platform,
                                               "sensor"        : sensor,
                                               "tag"           : tag,
                                               "start_day"     : start.strftime('%Y%m%d'),
                                               "start_time"    : start.strftime('%H%M')+'00',
                                               "end_time"      : end.strftime('%H%M')+'00',
                                               "creation_time" : datetime.utcnow().strftime("%Y%m%d%H%M%S")}
    out_dir = '/fjord/jgs/personal/mhiley/iff-zara-out-%s-%s/' % (platform, start.strftime('%Y'))
    if not os.path.exists(out_dir):
        # make directory, but NOT recursively
        os.mkdir(out_dir)
    existing = glob(out_dir+out_name[:-18]+'*.hdf')
    if len(existing) > 0:
        print "File for this time already exists %s" % out_name
    else:
        if zara:
            make_one_iff_zara(platform, start.strftime('%Y%m%d'), start.strftime('%H%M')+'00', end.strftime('%H%M')+'00', hirs, avhrr, out_dir)
        else:
            print "Writing %s" % out_name
            try:
                # mjh 2014/08/21: not sure why the hell I thought this entire thing would be "multiprocessing-safe".
                # let's try just calling out to the command line to see if that fixes our "tiny output files" problem.
                cmd = "cd %s ; " % out_dir
                cmd = cmd+"iff %s avhrr nss %s %s %s %s %s" % (platform, start.strftime('%Y%m%d'), start.strftime('%H%M')+'00', end.strftime('%H%M')+'00', avhrr, hirs)
                check_call(cmd, shell=True)
                #agg = aggregator.Aggregator(platform, AVHRR, tag, filelist, start, end, bowtie)
                #aggregator.write_agg_IFF_h4(agg, out_dir+out_name)
            except CalledProcessError:
                print "Error on %s" % out_name
    

def run_one_year(year, satid):
    """Run an entire year given a year and integer sat id."""
    from os.path import isdir
    satname = satid_map[satid][0]
    sat_2letter = satid_map[satid][1]

    hirs_dir = hirs_base_dir+satname+'/'+str(year)+'/'
    avhrr_dir = avhrr_base_dir+str(year)+'/level2/'
    if not isdir(hirs_dir):
        raise ValueError("hirs_dir doesn't exist")
    if not isdir(avhrr_dir):
        raise ValueError("avhrr_dir doesn't exist")

    hirs_files = glob(hirs_dir+'NSS.HIRX.'+sat_2letter+'.*hdf')
    # fix for noaa-18 2007:  there are a handful of *.gz.hdf files screwing things up
    for thing in hirs_files:
        if thing.endswith('gz.hdf'):
            hirs_files.remove(thing)
    avhrr_files = glob(avhrr_dir+'NSS.GHRR.'+sat_2letter+'.*.level2.hdf')
    #hirs_files = glob(hirs_dir+'NSS.HIRX.'+sat_2letter+'*D10001.S0325.E0520.B2378788.WI.hdf')
    #avhrr_files = glob(avhrr_dir+'NSS.GHRR.'+sat_2letter+'*D10001.S0325.E0520.B2378788.WI.level2.hdf')
    # assemble list of date/times for both lists
    hirs_times = make_date_set(hirs_files)
    avhrr_times = make_date_set(avhrr_files)
    # get times that exist in both sets
    time_set = hirs_times.intersection(avhrr_times)
    #for time in time_set:
    #    make_one_iff(time, satid, hirs_dir, avhrr_dir, zara=False)
    return (time_set, satid, hirs_dir, avhrr_dir)


if __name__ == "__main__":
    """Run one year given cmd line args: year and satid(int)"""
    import sys
    if len(sys.argv) < 3:
        raise ValueError("not enough args")
    satid = int(sys.argv[2])
    year = int(sys.argv[1])
    if not satid in satid_map:
        raise ValueError("invalid satellite identifier")
    if year < 1978 or year > 2020:
        raise ValueError("invalid year")
    run_one_year(year, satid)
