#!/usr/bin/env python
# encoding: utf-8
"""
Author: Nick Bearson, nickb@ssec.wisc.edu
Copyright (c) 2012 University of Wisconsin SSEC. All rights reserved.
"""

from sensor import Sensor
from descriptors import reify

from datetime import datetime, timedelta
import re

import h5py
import numpy as np

import logging
log = logging.getLogger(__name__)

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

from expressions import VIIRS_GEO_REGEX, VIIRS_IMG_REGEX, VIIRS_CMA_REGEX, search_list_regex

from expressions import CRIS_GEO_REGEX, CRIS_IMG_REGEX

# In microns, from:
# https://docs.google.com/viewer?a=v&q=cache:n8W6GmlTopcJ:cimss.ssec.wisc.edu/itwg/itsc/itsc13/session10/Puschell%2520VIIRS%2520Sensor%2520Performance%2520ITSC%2520FINAL.pdf+&hl=en&gl=us&pid=bl&srcid=ADGEESgz7OAWXj5ewYIG5fhN31sAOEB9c-ojiFqDj5FxuBaLpExWdFZb_kb9z8gMaaqQRKvQahvOqscHehp2JoH9QhTfF_Vy6Un0aGnSMf7-qBTGz4Hc0UqWUUXEX5e6Xa1VJjQ7KgFm&sig=AHIEtbTpGakVeHq00KBYnySxzFp9LLJZDA
# FIXME: should double-check against a more official source

b2w = {
  'SVM01' : 0.412,
  'SVM02' : 0.445,
  'SVM03' : 0.488,
  'SVM04' : 0.555,
  'SVM05' : 0.672,
  'SVM06' : 0.746,
  'SVM07' : 0.865,
  'SVM08' : 1.24,
  'SVM09' : 1.378,
  'SVM10' : 1.61,
  'SVM11' : 2.25,
  'SVM12' : 3.70,
  'SVM13' : 4.05,
  'SVM14' : 8.55,
  'SVM15' : 10.763,
  'SVM16' : 12.013,

  'SVI01' : 0.640,
  'SVI02' : 0.865,
  'SVI03' : 1.61,
  'SVI04' : 3.74,
  'SVI05' : 11.450,

  'SVDNB' : 0.7, # FIXME: what is the central wavelength for the DNB?
}


groups = {
  'SVM01' : 'All_Data/VIIRS-M1-SDR_All',
  'SVM02' : 'All_Data/VIIRS-M2-SDR_All',
  'SVM03' : 'All_Data/VIIRS-M3-SDR_All',
  'SVM04' : 'All_Data/VIIRS-M4-SDR_All',
  'SVM05' : 'All_Data/VIIRS-M5-SDR_All',
  'SVM06' : 'All_Data/VIIRS-M6-SDR_All',
  'SVM07' : 'All_Data/VIIRS-M7-SDR_All',
  'SVM08' : 'All_Data/VIIRS-M8-SDR_All',
  'SVM09' : 'All_Data/VIIRS-M9-SDR_All',
  'SVM10' : 'All_Data/VIIRS-M10-SDR_All',
  'SVM11' : 'All_Data/VIIRS-M11-SDR_All',
  'SVM12' : 'All_Data/VIIRS-M12-SDR_All',
  'SVM13' : 'All_Data/VIIRS-M13-SDR_All',
  'SVM14' : 'All_Data/VIIRS-M14-SDR_All',
  'SVM15' : 'All_Data/VIIRS-M15-SDR_All',
  'SVM16' : 'All_Data/VIIRS-M16-SDR_All',

  'SVI01' : 'All_Data/VIIRS-I1-SDR_All',
  'SVI02' : 'All_Data/VIIRS-I2-SDR_All',
  'SVI03' : 'All_Data/VIIRS-I3-SDR_All',
  'SVI04' : 'All_Data/VIIRS-I4-SDR_All',
  'SVI05' : 'All_Data/VIIRS-I5-SDR_All',

  'SVDNB' : 'All_Data/VIIRS-DNB-SDR_All',

  'GDNBO' : 'All_Data/VIIRS-DNB-GEO_All',
  'GITCO' : 'All_Data/VIIRS-IMG-GEO-TC_All',
  'GMTCO' : 'All_Data/VIIRS-MOD-GEO-TC_All',

  'IICMO' : 'All_Data/VIIRS-CM-IP_All',
}

refs_m_tags = ['SVM01', 'SVM02', 'SVM03', 'SVM04', 'SVM05', 'SVM06', 'SVM07',
               'SVM08', 'SVM09', 'SVM10', 'SVM11']

refs_i_tags = ['SVI01', 'SVI02', 'SVI03']

emis_m_tags = ['SVM12', 'SVM13', 'SVM14', 'SVM15', 'SVM16']

emis_i_tags = ['SVI04', 'SVI05']

bt_m_tags = ['SVM12', 'SVM13', 'SVM14', 'SVM15', 'SVM16']

bt_i_tags = ['SVI04', 'SVI05']

cm_tags = ['IICMO']

#FIXME: need a DNB tag too: 'SVDNB'

cris_centers = np.array([11030, 12020, 13335, 13635, 13935, 14235], dtype=np.float32) / 1000 # FIXME: these are taken from modis

cris_bands = [31, 32, 33, 34, 35, 36]  # FIXME: tag these as cris bands somehow


class VIIRS(Sensor):
  """
  This is a basic stub class for all of the functions a sensor should provide.  They are just a simplistic template, 
  and return NotImplementedError exceptions in case a subclass should forget to implement them.

  Takes a filename
  """

  SENSOR_EPOCH = datetime(1958, 1, 1)
  GROUPING_REGEX = '.*G[M|I]TCO_npp_(d[0-9]{8}_t[0-9]{7}_e[0-9]{7})_.*.h5'
  GROUPING_GLOB = '.*_npp_%s_.*.h5'

  def __init__(self, platform, tag, *args, **kwargs):
    # At the moment we have no extra args, but process them like this if we do...
    try:
      pass
#      self._w = kwargs.pop('w')
    except KeyError:
      pass

    super(VIIRS, self).__init__(*args, **kwargs)
#    print self._file_list

    self.SENSOR = "viirs"
    self.PLATFORM = platform
    self.tag = tag

    if self.tag == "svm":
      self.nDETECTORS = 16
    elif self.tag == "svi":
      self.nDETECTORS = 32
    elif self.tag == "dnb":
      raise NotImplementedError
    elif self.tag == "cma":
      self.nDETECTORS = 16
    else:
      raise NotImplementedError

    log.debug(self._file_list)

    self._geo_files = search_list_regex(self._file_list, VIIRS_GEO_REGEX)
    if len(self._geo_files) != 1:
      log.error("VIIRS object must be provided ONE geolocation file!  It got:")
      log.error(self._geo_files)
      raise RuntimeError, "VIIRS object must be provided ONE geolocation file!"

    self._img_files = search_list_regex(self._file_list, VIIRS_IMG_REGEX)
#    self._img_files = sorted(self._img_files)

    if self.tag in ['svm', 'svi', 'dnb']:
      if len(self._img_files) < 1:
        log.error("No filenames match:")
        log.error(VIIRS_IMG_REGEX)
        raise RuntimeError, "No img files provided!"

    self._cma_files = search_list_regex(self._file_list, VIIRS_CMA_REGEX)

    self._cris_geo_files = sorted(search_list_regex(self._file_list, CRIS_GEO_REGEX)) # FIXME: sorting to make them line up better - should probably be linked in some other way
    self._cris_img_files = sorted(search_list_regex(self._file_list, CRIS_IMG_REGEX))

    if self.tag in ['cma']:
      if len(self._cma_files) < 1:
        log.error("No filenames match:")
        log.error(VIIRS_CMA_REGEX)
        raise RuntimeError, "No img files provided!"


  @staticmethod
  def tag_from_fname(fname):
    for r in [VIIRS_IMG_REGEX, VIIRS_GEO_REGEX, VIIRS_CMA_REGEX]:
      m = re.match(r, fname)
      if m:
        return m.group(1)
    return []
  

  @staticmethod
  def date_bounds_from_fname(fname):
    DATE_REGEX = ".*npp_d([0-9]{8})_t([0-9]{7})_e([0-9]{7})_.*.h5"
    m = re.match(DATE_REGEX, fname)
    if m:
      d, t, e = m.group(1), m.group(2), m.group(3)
      frmt = '%Y%m%d%H%M%S%f'
      start = datetime.strptime(d+t, frmt)
      end = datetime.strptime(d+e, frmt)
      if end < start:
        end = end + timedelta(days=1)
      if end < start:
        raise RuntimeError, "can't get valid date bounds from file:\n%s\ndoesn't fit format:\n%s" % (fname, DATE_REGEX)
      return start, end
    else:
      raise RuntimeError, "can't get valid date bounds from file:\n%s\ndoesn't fit format:\n%s" % (fname, DATE_REGEX)


  # Generic function used by both ref_info and emis_info since they're both so similar
  def _generic_info(self, tag_list, data_name, fill_function):
    b_data = []
    b_wavs = []
    b_nums = []

    for tag in tag_list:

      files = [img for img in self._img_files if tag in img] # Get a list of all files that contain this tag

      data = None

      if len(files) > 1:
        log.error("Multiple files given to VIIRS object match %s" % tag)
        log.error(files)
        raise RuntimeError

      elif len(files) == 0:
        log.info("No files given to VIIRS object match %s, we'll try to use fills instead..." % tag)

      else:
        f = h5py.File(files[0])
        data = f[groups[tag] + "/" + data_name][:]
        try:
          factors = f[groups[tag] + "/" + data_name + "Factors"][:]
          invalid_mask = fill_function(data)
          invalids = data[invalid_mask]

          data = data * factors[0] + factors[1]
          data[invalid_mask] = invalids
        except:
          log.warn("Could not scale %s data for %s.  Most likely this is because it is already scaled." % (data_name, tag))
          pass

#      log.debug("Appending data for %s:" % (tag))
#      log.debug(data)
      b_data.append(data)
      b_wavs.append(b2w[tag])
      b_nums.append(tag)

    # Now go through get the shape of our valid arrays, make sure they're the same, etc.
    shape = None

    for d in b_data:
      if d is not None:
        if shape is None:
          shape = d.shape

        # Check that all of our arrays are the same shape
        else:
          if not shape == d.shape:
            log.error("Shape mismatch between VIIRS %s arrays." % (data_name))
            log.error("%s vs. %s" % (shape, d.shape))
            raise RuntimeError

    # If we get through all that and shape is still None (ie: no data for anything), we use latitude's shape
    # FIXME: what should we do in the case that a VIIRS object is initialized with no img data and no geo data?
    if shape is None:
      shape = self.Latitude.shape

    filled = np.ones(shape) * -999
    for i, d in enumerate(b_data):
      if d is None:
        log.info("Filling %s with fill values." % (b_nums[i]))
        b_data[i] = filled

    all_data = np.array(b_data, dtype=np.float32)
    all_wavs = np.array(b_wavs, dtype=np.float32)
    all_nums = np.array(b_nums)

    return (all_data, all_wavs, all_nums)


  @reify
  def _ref_info(self):
    def ref_get_invalids(data): return (data >= 65528)
    refs_tags = []
    if self.tag == "svm":
      refs_tags = refs_m_tags
    elif self.tag == "svi":
      refs_tags = refs_i_tags
    elif self.tag == "dnb":
      refs_tags = refs_dnb_tags
    return self._generic_info(refs_tags, "Reflectance", ref_get_invalids)


  @reify
  def _emis_info(self):
    def emis_get_invalids(data): return (data <= -999)
    emis_tags = []
    if self.tag == "svm":
      emis_tags = emis_m_tags
    elif self.tag == "svi":
      emis_tags = emis_i_tags
    elif self.tag == "dnb":
      emis_tags = emis_dnb_tags

    data, wavs, nums = self._generic_info(emis_tags, "Radiance", emis_get_invalids)
    
    # Now add in CrIS!

    return (data, wavs, nums)

  @reify
  def _bt_info(self):
    def bt_get_invalids(data): return (data >= 65528)
    emis_tags = []
    if self.tag == "svm":
      emis_tags = bt_m_tags
    elif self.tag == "svi":
      emis_tags = bt_i_tags
    elif self.tag == "dnb":
      emis_tags = bt_dnb_tags
    return self._generic_info(emis_tags, "BrightnessTemperature", bt_get_invalids)

  @reify
  def ReflectiveSolarBands(self):
    return self._ref_info[0]

  @reify
  def ReflectiveBandCenters(self):
    return self._ref_info[1]

  @reify
  def ReflectiveBandNums(self):
    return self._ref_info[2]


  @reify
  def CrisBands(self):
    import c2v
    log.debug("making cris bands!")

    crisshape = (len(cris_bands),) + self.Latitude.shape
    allbts  = np.empty(crisshape)
    allbts.fill(-999.0)
    allrads = np.empty(crisshape)
    allrads.fill(-999.0)

    for i in range(len(self._cris_geo_files)):
      cris_geo = self._cris_geo_files[i]
      cris_img = self._cris_img_files[i]
      viirs_geo = self._geo_files[0]

      log.debug("CrisBands processing...")
      log.debug("cris geo file: %s" % (cris_geo))
      log.debug("cris img file: %s" % (cris_img))
      log.debug("viirs geo file: %s" % (viirs_geo))

      mapping = c2v.cris2viirs_transform(cris_geo, viirs_geo)

      radlist = []
      btlist = []

      for band in cris_bands:
        log.debug("adding cris band %d" % (band))
        brads, bbts = c2v.cris2modis_translate(cris_img, band)
        vrads = c2v.apply_mapping(mapping, brads)
        vbts = c2v.apply_mapping(mapping, bbts)
        radlist.append(vrads)
        btlist.append(vbts)

      rads = np.array(radlist, dtype=radlist[0].dtype)
      bts = np.array(btlist, dtype=btlist[0].dtype)
    
       
      # mix the new array into the total array - FIXME: REMEMBER WE'RE USING NEAREST - ADD WEIGHTING WHEN ITS AVAILABLE
      def combinate(total, new):
        if total == None:
          log.debug("processing cris: first in set")
          total = new
        else:
          assert( total.shape == new.shape ) # new should always be the same shape as total... right? catch if it's not because that would break other things
          total[total == -999.0] = new[total == -999.0] # FIXME: -999.0 is just our fill - replace this with checking for larger weights?
        return total

      
      allrads = combinate(allrads, rads)
      allbts = combinate(allbts, bts)

    log.debug("returning from CrisBands")
    return allrads, allbts



  def SounderEmissive(self):
    return self.CrisBands[0]

  def SounderBTs(self):
    return self.CrisBands[1]

  @reify
  def SounderCenters(self):
    return cris_centers

  @reify
  def SounderNums(self):
    clist = []
    for band in cris_bands:
      clist.append("CrIS-%d" % (band))
    return clist

  @reify
  def EmissiveBands(self):
    return self._emis_info[0]

  @reify
  def EmissiveBandCenters(self):
    return self._emis_info[1]

  @reify
  def EmissiveBandNums(self):
    return self._emis_info[2]

  @reify
  def BrightnessTemperatureBands(self):
    return self._bt_info[0]

  @reify
  def BrightnessTemperatureBandCenters(self):
    return self._bt_info[1]

  @reify
  def BrightnessTemperatureBandNums(self):
    return self._bt_info[2]



  @reify
  def Height(self):
    geo_fname = self._geo_files[0]
    f = h5py.File(geo_fname)
    tag = self.tag_from_fname(geo_fname)
    height = f[groups[tag] + '/Height'][:]
    return height[:].astype(np.float32)


  @reify
  def Range(self):
    geo_fname = self._geo_files[0]
    f = h5py.File(geo_fname)
    tag = self.tag_from_fname(geo_fname)
    r = f[groups[tag] + '/SatelliteRange'][:]
    return r.astype(np.float32)


  @reify
  def SensorZenith(self):
    geo_fname = self._geo_files[0]
    f = h5py.File(geo_fname)
    tag = self.tag_from_fname(geo_fname)
    r = f[groups[tag] + '/SatelliteZenithAngle'][:]
    return r.astype(np.float32)


  @reify
  def SensorAzimuth(self):
    geo_fname = self._geo_files[0]
    f = h5py.File(geo_fname)
    tag = self.tag_from_fname(geo_fname)
    r = f[groups[tag] + '/SatelliteAzimuthAngle'][:]
    return r.astype(np.float32)


  @reify
  def SolarZenith(self):
    geo_fname = self._geo_files[0]
    f = h5py.File(geo_fname)
    tag = self.tag_from_fname(geo_fname)
    r = f[groups[tag] + '/SolarZenithAngle'][:]
    return r.astype(np.float32)


  @reify
  def SolarAzimuth(self):
    geo_fname = self._geo_files[0]
    f = h5py.File(geo_fname)
    tag = self.tag_from_fname(geo_fname)
    r = f[groups[tag] + '/SolarAzimuthAngle'][:]
    return r.astype(np.float32)


  @reify
  def Latitude(self):
    geo_fname = self._geo_files[0]
    f = h5py.File(geo_fname)
    tag = self.tag_from_fname(geo_fname)
    r = f[groups[tag] + '/Latitude'][:]
    return r[:]

  @reify
  def Longitude(self):
    geo_fname = self._geo_files[0]
    f = h5py.File(geo_fname)
    tag = self.tag_from_fname(geo_fname)
    r = f[groups[tag] + '/Longitude'][:]
    return r[:]


  @reify
  def ScanStartTime(self):
    geo_fname = self._geo_files[0]
    f = h5py.File(geo_fname)
    tag = self.tag_from_fname(geo_fname)
    r = f[groups[tag] + '/StartTime'][:]
    r = r / 1000000.0  # convert to seconds
    return r[:]
